#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "p2lassert.h"
#include "c3dsmesh.h"
#include <intmd/sceneintmd.h>



float fRnd() { return ((float) (((unsigned int) rand()) & 0xffff)) / ((unsigned int) 0xffff); }

bool C3dsMesh::SetupMesh(CDmaVif1 * pDmaVif1, CMeshIntmd * pMesh)
{
	// num materials	
	int num_surfaces = pMesh->GetNumSurfaces();


	// entirely hacked area for results Begin
	SRenderGadget * pAllStrips = NULL; 
	int32 iTotalNumOfStrips = 0, iCurStrip = 0;
	for(int s=0; s<num_surfaces; s++)
		iTotalNumOfStrips += pMesh->GetSurface(s).GetNumStrips();
	pAllStrips = new SRenderGadget[iTotalNumOfStrips];

	for(int s=0; s<num_surfaces; s++)
	{
		CSurfaceIntmd &surf = pMesh->GetSurface(s);

		// strips
		for(int k=0; k<surf.GetNumStrips(); k++)
		{
			// fetch amount of vertices in current strip
			int32 verts_in_strip = surf.GetStrip(k).GetArrayLen();

			// allocate data for strip
			SRenderGadget &gadget = pAllStrips[iCurStrip++];
			
			gadget.m_iNumElems = verts_in_strip;
			gadget.m_iPrimType = PRIMTYPE_STRIP;
			
			gadget.m_pVerts = new float[4*verts_in_strip];
			gadget.m_pCols = new int[4*verts_in_strip];
			gadget.m_pTexCrds = new float[4*verts_in_strip];

			int r = (int) (fRnd()*127);
			int g = (int) (fRnd()*127);
			int b = (int) (fRnd()*127);

			for(int i=0; i<verts_in_strip; i++)
			{
				const int32 index = surf.GetStrip(k).GetIndex(i);

				// insert vertex and color
				gadget.m_pVerts[i*4+0] = surf.GetVertex(index).x;
				gadget.m_pVerts[i*4+1] = surf.GetVertex(index).y;
				gadget.m_pVerts[i*4+2] = surf.GetVertex(index).z;

				gadget.m_pCols[i*4+0] = r;
				gadget.m_pCols[i*4+1] = g;
				gadget.m_pCols[i*4+2] = b;
				gadget.m_pCols[i*4+3] = 0;

				// do texture coordinate
				gadget.m_pTexCrds[i*4+0] = surf.GetST(index).x;
				gadget.m_pTexCrds[i*4+1] = surf.GetST(index).y;

				// insert normal
				gadget.m_pTexCrds[i*4+2] = surf.GetNormal(index).x;
				gadget.m_pTexCrds[i*4+3] = surf.GetNormal(index).y;
				gadget.m_pVerts[i*4+3] = surf.GetNormal(index).z;
			}
		}
	}

	// convert strips
	bool bRes = ConvertToChunk(pDmaVif1, pAllStrips, iTotalNumOfStrips);
	CleanStrips(pAllStrips, iTotalNumOfStrips);

	return bRes;
}

C3dsMesh::C3dsMesh()
{

}


void C3dsMesh::CleanUp()
{
	IMesh::CleanUp();
}

void C3dsMesh::CleanStrips(SRenderGadget * pStrips, int iTotStrips)
{
	if(pStrips != NULL)
	{
		for(int i=0; i<iTotStrips; i++)
		{
			if(pStrips[i].m_pVerts != NULL)
				delete [] pStrips[i].m_pVerts;

			if(pStrips[i].m_pCols != NULL)
				delete [] pStrips[i].m_pCols;

			if(pStrips[i].m_pTexCrds != NULL)
				delete [] pStrips[i].m_pTexCrds;
		}

		delete pStrips;
	}
}
