#include "imesh.h"
#include <stdio.h>
#include "macro_codes.h"
#include <math.h>
#include "dmavif1.h"
#include "imeshinstance.h"
#include "stitchhack.h"
#include "p2lassert.h"


// Giftags used for the rendering process
const uint64 orgGIFTagStrip[2] = 
{ 
	GIFTAG(0, 0, 1, PRIM_SET(4, 1, 0, 0, 0, 0, 0, 0, 0), 0, 3), 
	GIFTAG_REGS4(2, 1, 4, 0)
};
   
const uint64 orgGIFTagTriangle[2] =
{ 
	GIFTAG(0, 0, 1, PRIM_SET(3, 1, 0, 0, 0, 0, 0, 0, 0), 0, 9), 
	GIFTAG_REGS12(2, 1, 4, 2, 1, 4, 2, 1, 4, 0, 0, 0)
};




IMeshInstance * IMesh::CreateInstance(CDmaVif1 * pInstDmaVif1)
{
	IMeshInstance * pMeshInst = NULL;

	// Has a mesh been built yet?
	if(m_iNumOfChunks <= 0)
		return NULL;

	if((pMeshInst=new IMeshInstance(this, pInstDmaVif1)) == NULL) 
		return NULL;
	else
		return pMeshInst;
}

char SnapToChar(int a)
{
	float fTmp;
	if(a<0) fTmp = a-0.5f;
	else fTmp = a+0.5f;

	if(fTmp>0x7f) fTmp = 0x7f;
	else if(fTmp < -0x7f) fTmp = -0x7f;

	return (char) fTmp;
}

int SnapScale(float fVal, float fScale)
{
	float fRes = fVal * fScale;
	if(fRes<0) fRes-=0.5f;
	else fRes+=0.5f;

	if(fRes<0 && (-fRes)>0x7fffffffL) fRes = -0x7fffffffL;
	else if(fRes>=0 && fRes>0x7fffffffL) fRes = 0x7fffffffL;

	return (int) fRes;
}

bool IMesh::ConvertToChunk(CDmaVif1 * pDmaVif1, SRenderGadget * pAllStrips, int iTotalNumOfStrips)
{
	CleanUp();

	// setup
	pDmaVif1->Reset();
	m_pDmaTags = pDmaVif1->GetCurLocation();

	// calculate number of chunks needed
	m_iNumOfChunks = 1;	// Initial value
	int iStripNumInChunk = 0;
	int iNumElem = 0;

	for(int iC=0; iC<iTotalNumOfStrips; iC++)
	{
		iNumElem += pAllStrips[iC].m_iNumElems;
		++iStripNumInChunk;

		if(iStripNumInChunk>16 || iNumElem>75)	// Didn't find space in chunk
		{
			++m_iNumOfChunks;	// move on to next chunk
			iStripNumInChunk = 1;
			iNumElem = pAllStrips[iC].m_iNumElems;
		}
	}

	// Alloc Chunks
	m_pChunks = NULL;

	if((m_pChunks=new SStripChunk[m_iNumOfChunks])==NULL)
	{ CleanUp(); return false; }


	// Determine amount of strips for each chunk
	int iStrip = 0;

	for(int iCHunk=0; iCHunk<m_iNumOfChunks; iCHunk++)
	{
		SStripChunk &Chnk = m_pChunks[iCHunk];
		Chnk.m_iNumOfStrips = 0;
		Chnk.m_pNumVertsInStrip = NULL;
		Chnk.m_pVerts = NULL;
		Chnk.m_pCols = NULL;
		Chnk.m_pTexCrds = NULL;
		//Chnk.m_uGifTagStrip = *((u_long128 *) &orgGIFTagStrip[0]);
		//Chnk.m_uGifTagTriangle = *((u_long128 *) &orgGIFTagTriangle[0]);
		((uint64 *) &Chnk.m_uGifTagStrip)[0] = orgGIFTagStrip[0];
		((uint64 *) &Chnk.m_uGifTagStrip)[1] = orgGIFTagStrip[1];

		((uint64 *) &Chnk.m_uGifTagTriangle)[0] = orgGIFTagTriangle[0];
		((uint64 *) &Chnk.m_uGifTagTriangle)[1] = orgGIFTagTriangle[1];


		iNumElem = 0;

		while((1+Chnk.m_iNumOfStrips)<=16 && (iNumElem+pAllStrips[iStrip].m_iNumElems)<=75 && iStrip<iTotalNumOfStrips)
		{
			iNumElem += pAllStrips[iStrip].m_iNumElems;
			++Chnk.m_iNumOfStrips;
			++iStrip;
		}
	
		int iSizeInQuads = Chnk.m_iNumOfStrips>>2;	// was zero
		if((Chnk.m_iNumOfStrips&0x3)!=0)
			++iSizeInQuads;

		if((Chnk.m_pNumVertsInStrip=(int *) (new uint128[iSizeInQuads])) == NULL)
		{ CleanUp(); return false; }
	}

	assert(iStrip == iTotalNumOfStrips);

	// Get Strips
	iStrip = 0;
	float fMaxDeltaSizeVert = 0;
	float fMaxDeltaSizeST 	= 0;
	for(int iCHunk=0; iCHunk<m_iNumOfChunks; iCHunk++)
	{
		SStripChunk &Chnk = m_pChunks[iCHunk];
		Chnk.m_iNumElemsInChunk = 0;
		Chnk.m_pVerts = NULL; 
		Chnk.m_pCols = NULL;
		Chnk.m_pTexCrds = NULL;

		for(int iS=0; iS<Chnk.m_iNumOfStrips; iS++)
		{
			Chnk.m_iNumElemsInChunk += pAllStrips[iStrip+iS].m_iNumElems;
			Chnk.m_pNumVertsInStrip[iS] = pAllStrips[iStrip+iS].m_iNumElems;	// Was zero
		}

		if((Chnk.m_pVerts=new float[4*Chnk.m_iNumElemsInChunk]) == NULL)
		{ CleanUp(); return false; }

		if((Chnk.m_pCols=new int[4*Chnk.m_iNumElemsInChunk]) == NULL)
		{ CleanUp(); return false; }

		if((Chnk.m_pTexCrds=new float[4*Chnk.m_iNumElemsInChunk]) == NULL)
		{ CleanUp(); return false; }

		// Fill up with vertices
		int iElem = 0;
		//printf("numS: %d\n", Chnk.m_iNumOfStrips);
		//printf("numE: %d\n", Chnk.m_iNumElemsInChunk);
		for(int iS=0; iS<Chnk.m_iNumOfStrips; iS++)
		{
			SRenderGadget &sCurStrip = pAllStrips[iStrip+iS];
			

			for(int iCnt=0; iCnt<sCurStrip.m_iNumElems; iCnt++)
			{
				((uint128 *) Chnk.m_pVerts)[iElem] = ((uint128 *) sCurStrip.m_pVerts)[iCnt];
				((uint128 *) Chnk.m_pCols)[iElem] = ((uint128 *) sCurStrip.m_pCols)[iCnt];
				((uint128 *) Chnk.m_pTexCrds)[iElem] = ((uint128 *) sCurStrip.m_pTexCrds)[iCnt];

				if(iCnt > 0)
				{
					for(int iT=0; iT<3; iT++)
					{
						float fAbsDeltaVert = fabs(sCurStrip.m_pVerts[4*iCnt + iT] - sCurStrip.m_pVerts[4*(iCnt-1) + iT]);
						float fAbsDeltaST	  = fabs(sCurStrip.m_pTexCrds[4*iCnt + iT] - sCurStrip.m_pTexCrds[4*(iCnt-1) + iT]);
						
						if(fAbsDeltaVert > fMaxDeltaSizeVert) fMaxDeltaSizeVert = fAbsDeltaVert;
						if(iT<2 && fAbsDeltaST > fMaxDeltaSizeST) fMaxDeltaSizeST = fAbsDeltaST;
					}
				}
				
				++iElem;
			}
		}

		iStrip += Chnk.m_iNumOfStrips;
	}
	
	// Pack strips to prepare for difference mode
	// (lookup difference mode in EE Users, vif section)
	for(int iCHunk=0; iCHunk<m_iNumOfChunks; iCHunk++)
	{
		SStripChunk &Chnk = m_pChunks[iCHunk];

		// locate memory for current chunk
		uint32 * pDstVerts = (uint32 *) Chnk.m_pVerts;
		uint32 * pDstSTs = (uint32 *) Chnk.m_pTexCrds;
		uint32 * pDstCols = (uint32 *) Chnk.m_pCols;

		// downscale to achieve best fit for delta chars
		float fVScale = 126.0f / fMaxDeltaSizeVert;
		float fSTScale = 126.0f / fMaxDeltaSizeST;
		float * pVSrcData = Chnk.m_pVerts;
		float * pSTSrcData = Chnk.m_pTexCrds;
		uint32 * pColSrcData = (uint32 *) Chnk.m_pCols;
		
		for(int iS=0; iS<Chnk.m_iNumOfStrips; iS++)
		{
			int iNumVertsInStrip = Chnk.m_pNumVertsInStrip[iS];

			// necessary to backup 5 vals to use same buffer for dstmem
			float fRecVx = pVSrcData[0], fRecVy = pVSrcData[1];
			float fRecVz = pVSrcData[2], fRecVw = pVSrcData[3];
			float fRecVx2 = pVSrcData[4];

			// first four Vert longs for ROW register
			*pDstVerts++ = SnapScale(pVSrcData[0], fVScale);
			*pDstVerts++ = SnapScale(pVSrcData[1], fVScale);
			*pDstVerts++ = SnapScale(pVSrcData[2], fVScale);
			*pDstVerts++ = SnapScale(pVSrcData[3], 63);	// nz

			// necessary to backup 5 vals to use same buffer for dstmem
			float fRecSTx = pSTSrcData[0], fRecSTy = pSTSrcData[1];
			float fRecSTz = pSTSrcData[2], fRecSTw = pSTSrcData[3];
			float fRecSTx2 = pSTSrcData[4];

			// first four ST longs for ROW register
			*pDstSTs++ = SnapScale(pSTSrcData[0], fSTScale);
			*pDstSTs++ = SnapScale(pSTSrcData[1], fSTScale);
			*pDstSTs++ = SnapScale(pSTSrcData[2], 63);	// nx
			*pDstSTs++ = SnapScale(pSTSrcData[3], 63);	// ny

			char cX1=0, cY1=0, cZ1=0, cW1=0;
			char cX2=0, cY2=0, cZ2=0, cW2=0;
			for(int i=0; i<iNumVertsInStrip; i++)
			{
				// Fill up current vertex
				((char *) pDstVerts)[0] = cX1;
				((char *) pDstVerts)[1] = cY1;
				((char *) pDstVerts)[2] = cZ1;
				((char *) pDstVerts)[3] = cW1;
				++pDstVerts;

				// Fill up current ST
				((char *) pDstSTs)[0] = cX2;
				((char *) pDstSTs)[1] = cY2;
				((char *) pDstSTs)[2] = cZ2;
				((char *) pDstSTs)[3] = cW2;
				++pDstSTs;

				// Fill up Colors
				((unsigned char *) pDstCols)[0] = (unsigned char) pColSrcData[0];
				((unsigned char *) pDstCols)[1] = (unsigned char) pColSrcData[1];
				((unsigned char *) pDstCols)[2] = (unsigned char) pColSrcData[2];
				((unsigned char *) pDstCols)[3] = (unsigned char) pColSrcData[3];
				++pDstCols;
				
				

				// Calculate next
				if(i<iNumVertsInStrip-1)
				{
					cX1 = SnapToChar(SnapScale(fRecVx2, fVScale) - SnapScale(fRecVx, fVScale));
					cY1 = SnapToChar(SnapScale(pVSrcData[4+1], fVScale) - SnapScale(fRecVy, fVScale));
					cZ1 = SnapToChar(SnapScale(pVSrcData[4+2], fVScale) - SnapScale(fRecVz, fVScale));
					cW1 = SnapToChar(SnapScale(pVSrcData[4+3], 63) - SnapScale(fRecVw, 63));

					cX2 = SnapToChar(SnapScale(fRecSTx2, fSTScale) - SnapScale(fRecSTx, fSTScale));
					cY2 = SnapToChar(SnapScale(pSTSrcData[4+1], fSTScale) - SnapScale(fRecSTy, fSTScale));
					cZ2 = SnapToChar(SnapScale(pSTSrcData[4+2],63)-SnapScale(fRecSTz,63));
					cW2 = SnapToChar(SnapScale(pSTSrcData[4+3],63)-SnapScale(fRecSTw,63));	

					// keep backing up
					fRecVx = fRecVx2, fRecVy = pVSrcData[4+1];
					fRecVz = pVSrcData[4+2], fRecVw = pVSrcData[4+3];
					fRecVx2 = pVSrcData[4+4];

					fRecSTx = fRecSTx2, fRecSTy = pSTSrcData[4+1];
					fRecSTz = pSTSrcData[4+2], fRecSTw = pSTSrcData[4+3];
					fRecSTx2 = pSTSrcData[4+4];
				}

				pVSrcData += 4; pSTSrcData += 4; pColSrcData += 4;
			}
		}
	}

	////////// setup DmaChain ///////////
	/////////////////////////////////////
	
	InitChain(pDmaVif1, STCYCL(1,1));
	AddWordToChain(pDmaVif1, UNPACK_TO_TOP(V4_32, 2, PAD_CHUNK+16));

	// add strip (GifTag)
	AddWordToChain(pDmaVif1, ((uint32 *) orgGIFTagStrip)[0]);
	AddWordToChain(pDmaVif1, ((uint32 *) orgGIFTagStrip)[1]);
	AddWordToChain(pDmaVif1, ((uint32 *) orgGIFTagStrip)[2]);
	AddWordToChain(pDmaVif1, ((uint32 *) orgGIFTagStrip)[3]);
	
	// add triangle (GifTag)
	AddWordToChain(pDmaVif1, ((uint32 *) orgGIFTagTriangle)[0]);
	AddWordToChain(pDmaVif1, ((uint32 *) orgGIFTagTriangle)[1]);
	AddWordToChain(pDmaVif1, ((uint32 *) orgGIFTagTriangle)[2]);
	AddWordToChain(pDmaVif1, ((uint32 *) orgGIFTagTriangle)[3]);
	
	// Insert init regs call
	AddWordToChain(pDmaVif1, MSCALL(2)); 

	// add all chunks
	for(int iCount=0; iCount<m_iNumOfChunks; iCount++)
	{
		SStripChunk &Chnk = m_pChunks[iCount];

		assert(Chnk.m_iNumElemsInChunk<=75);
		assert(Chnk.m_iNumOfStrips<=16);

		// add vertices
		AddWordToChain(pDmaVif1, STCYCL(1,2));
		AddWordToChain(pDmaVif1, STMOD(2));
		uint32 * pVPtr = (uint32 *) Chnk.m_pVerts;
		uint32 uVertsPassed = 0;
		for(int iS=0; iS<Chnk.m_iNumOfStrips; iS++)
		{
			int iNumVertsInCurStrip = Chnk.m_pNumVertsInStrip[iS];

			AddWordToChain(pDmaVif1, STROW);
			AddWordToChain(pDmaVif1, *pVPtr++);
			AddWordToChain(pDmaVif1, *pVPtr++);
			AddWordToChain(pDmaVif1, *pVPtr++);
			AddWordToChain(pDmaVif1, *pVPtr++);
			uint32 uOffs = 0+PAD_CHUNK+Chnk.m_iNumElemsInChunk+(2*uVertsPassed);
			AddWordToChain(pDmaVif1, UNPACK_TO_TOP(V4_8, iNumVertsInCurStrip, uOffs) );
			
			for(int iN=0; iN<iNumVertsInCurStrip; iN++)
				AddWordToChain(pDmaVif1, *pVPtr++);

			// end
			uVertsPassed += iNumVertsInCurStrip;
		}
		
		// add TexCrds
		uint32 * pSTPtr = (uint32 *) Chnk.m_pTexCrds;
		uint32 uSTsPassed = 0;
		for(int iS=0; iS<Chnk.m_iNumOfStrips; iS++)
		{
			int iNumSTsInCurStrip = Chnk.m_pNumVertsInStrip[iS];
			
			AddWordToChain(pDmaVif1, STROW);
			AddWordToChain(pDmaVif1, *pSTPtr++);
			AddWordToChain(pDmaVif1, *pSTPtr++);
			AddWordToChain(pDmaVif1, *pSTPtr++);
			AddWordToChain(pDmaVif1, *pSTPtr++);
			
			uint32 uOffs = 1+PAD_CHUNK+Chnk.m_iNumElemsInChunk+(2*uSTsPassed);
			AddWordToChain(pDmaVif1, UNPACK_TO_TOP(V4_8, iNumSTsInCurStrip, uOffs) );
			
			for(int iN=0; iN<iNumSTsInCurStrip; iN++)
				AddWordToChain(pDmaVif1, *pSTPtr++);

			// end
			uSTsPassed += iNumSTsInCurStrip;
		}
		
		// Deliver Colours
		AddWordToChain(pDmaVif1, STMOD(0));
		AddWordToChain(pDmaVif1, STCYCL(1,1));
		AddWordToChain(pDmaVif1, UNPACK_TO_TOP_UNSIGNED(V4_8, Chnk.m_iNumElemsInChunk, PAD_CHUNK+0) );
		uint32 * pColPtr = (uint32 *) Chnk.m_pCols;
		for(int iCol=0; iCol<Chnk.m_iNumElemsInChunk; iCol++)
			AddWordToChain(pDmaVif1, *pColPtr++);
			
		// Deliver NumVertsInStrip
		uint32 uNumUnp = Chnk.m_iNumOfStrips;
		uint32 uMod4 = uNumUnp & 0x3;
		if(uMod4!=0) uNumUnp += (4-uMod4); 
		assert((uNumUnp&0x3)==0);
		AddWordToChain(pDmaVif1, UNPACK_TO_TOP(V4_32, uNumUnp/4, PAD_CHUNK+(Chnk.m_iNumElemsInChunk*3)) );
		uint32 * pNumsPtr = (uint32 *) Chnk.m_pNumVertsInStrip;
		for(uint32 uStr=0; uStr<uNumUnp; uStr++)
			if(uStr < (uint32) Chnk.m_iNumOfStrips)
				AddWordToChain(pDmaVif1, *pNumsPtr++);
			else
				AddWordToChain(pDmaVif1, 0);
		
		// Deliver num of strips in current chunk
		float fVscale = fMaxDeltaSizeVert / 126.0f;
		float fSTRecScale = (fMaxDeltaSizeST > 0) ? 126.0f / fMaxDeltaSizeST : 1; 
		AddWordToChain(pDmaVif1, UNPACK_TO_TOP(V4_32, 1, 0) );
		AddWordToChain(pDmaVif1, (uint32) Chnk.m_iNumElemsInChunk);
		AddWordToChain(pDmaVif1, (uint32) Chnk.m_iNumOfStrips);
		AddWordToChain(pDmaVif1, *((uint32 *) &fVscale));
		AddWordToChain(pDmaVif1, *((uint32 *) &fSTRecScale));

		// add draw call
		AddWordToChain(pDmaVif1, MSCALL(6) );
	}

	// Terminate (add a ret_tag)
	Terminate(pDmaVif1);

	return true;
}


void IMesh::CleanUp()
{
	if(m_pChunks!=NULL)
	{
		for(int iCnt=0; iCnt<m_iNumOfChunks; iCnt++)
		{
			if(m_pChunks[iCnt].m_pNumVertsInStrip!=NULL)
				delete [] m_pChunks[iCnt].m_pNumVertsInStrip;

			if(m_pChunks[iCnt].m_pVerts!=NULL)
				delete [] m_pChunks[iCnt].m_pVerts;

			if(m_pChunks[iCnt].m_pCols!=NULL)
				delete [] m_pChunks[iCnt].m_pCols;

			if(m_pChunks[iCnt].m_pTexCrds!=NULL)
				delete [] m_pChunks[iCnt].m_pTexCrds;
		}

		delete [] m_pChunks;
	}

	m_iNumOfChunks = 0;
}

IMesh::IMesh()
{
	m_iNumOfChunks = 0;
	m_pChunks = NULL;
}
